# -*- coding: utf-8 -*-
"""test_mnist.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1q11Vtg9xNj7JV8LDf1rYiwezTdV-MPRO
"""

!python --version
!for p in torch torchvision numpy; do pip list | grep "^$p[ \t]"; done

"""Python 3.10.12

torch                              2.5.1+cu121
torchvision                        0.20.1+cu121
numpy                              1.26.4
"""

import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import datasets, transforms
from torchvision.transforms import ToTensor
from torch.utils.data import TensorDataset

import numpy

# Get cpu or gpu device for training.
device = "cuda" if torch.cuda.is_available() else "cpu"
print(f"Using {device} device")

transform = transforms.ToTensor()
training_data_1 = datasets.MNIST(root='./data', train=True, download=True, transform=transform)
test_data_1 = datasets.MNIST(root='./data', train=False, download=True, transform=transform)

batch_size = 50

# Create data loaders.
train_dataloader_1 = DataLoader(training_data_1, batch_size=batch_size)
test_dataloader_1 = DataLoader(test_data_1, batch_size=batch_size)

labels = torch.LongTensor([])
for batch, (X, y) in enumerate(test_dataloader_1):
  labels = torch.cat((labels, y))
print(labels)

images = torch.FloatTensor([])
for batch, (X, y) in enumerate(test_dataloader_1):
  images = torch.cat((images, X))
images = images.reshape(10000, 784)
print(images.shape)

print("Loading labels ...")
# from labels import labels
# labels = torch.Tensor(labels)
labels = torch.nn.functional.one_hot(labels.long()) # .to(torch.float32)
# labels = labels.to(torch.long)
train_labels = labels[0:8000]
test_labels = labels[8000:9000]

print("Loading images ...")
# from images import images
train_images = images[0:8000]
test_images = images[8000:9000]

# train_images = [train_images]
# train_images = numpy.array(train_images).astype('float32') / 255
# train_images = torch.Tensor(train_images)
# train_images = train_images / 255.0

# test_images = [test_images]
# test_images = numpy.array(test_images).astype('float32') / 255
# test_images = torch.Tensor(test_images)
# test_images = test_images / 255.0

print(train_labels.shape)
print(test_labels.shape)

print(train_labels)

print(test_labels)

print(train_images.shape)

print(test_images.shape)

print(test_images[0])

training_data = TensorDataset(torch.tensor(train_images,dtype=torch.float32), torch.tensor(train_labels,dtype=torch.long))
test_data = TensorDataset(torch.tensor(test_images,dtype=torch.float32), torch.tensor(test_labels,dtype=torch.long))

batch_size = 8

# Create data loaders.
train_dataloader = DataLoader(training_data, batch_size=batch_size)
test_dataloader = DataLoader(test_data, batch_size=batch_size)

for X, y in test_dataloader:
    print(f"Shape of X [N, C, H, W]: {X.shape}")
    print(f"Shape of y: {y.shape} {y.dtype}")
    break

# Get cpu or gpu device for training.
device = "cuda" if torch.cuda.is_available() else "cpu"
print(f"Using {device} device")

# Define model
class NeuralNetwork(nn.Module):
    def __init__(self):
        super().__init__()
        # self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(28*28, 512),
            nn.ReLU(),
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.Linear(512, 10)
        )

    def forward(self, x):
        # x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits

model = NeuralNetwork().to(device)
print(model)

# loss_fn = nn.CrossEntropyLoss()

def loss_fn(output, target):
    loss = torch.sum((output - target) ** 2)
    return loss

optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)

def train(dataloader, model, loss_fn, optimizer):
    size = len(dataloader.dataset)
    model.train()
    for batch, (X, y) in enumerate(dataloader):
        X, y = X.to(device), y.to(device)

        # Compute prediction error
        pred = model(X)
        loss = loss_fn(pred, y)

        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if batch % 100 == 0:
            loss, current = loss.item(), batch * len(X)
            print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

def test(dataloader, model, loss_fn):
    size = len(dataloader.dataset)
    num_batches = len(dataloader)
    model.eval()
    test_loss, correct = 0, 0
    with torch.no_grad():
        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            pred = model(X)
            test_loss += loss_fn(pred, y).item()
            # y1 = torch.matmul(y, torch.LongTensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
            y1 = torch.matmul(y.float(), torch.FloatTensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).to(device)).long()
            correct += (pred.argmax(1) == y1).type(torch.float32).sum().item()
    test_loss /= num_batches
    correct /= size
    print(f"Test Error: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")

epochs = 10
for t in range(epochs):
    print(f"Epoch {t+1}\n-------------------------------")
    train(train_dataloader, model, loss_fn, optimizer)
    test(test_dataloader, model, loss_fn)
print("Done!")

torch.save(model.state_dict(), "model.pth")
print("Saved PyTorch Model State to model.pth")

model = NeuralNetwork()
model.load_state_dict(torch.load("model.pth"))

classes = [
    "Zero ",
    "One  ",
    "Two  ",
    "Three",
    "Four ",
    "Five ",
    "Six  ",
    "Seven",
    "Eight",
    "Nine ",
]

model.eval()

ngood = 0
nbad = 0

for i in range(100):
  x, y = test_data[i][0], test_data[i][1]
  with torch.no_grad():
    pred = model(x)
    y1 = torch.matmul(y, torch.LongTensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
    predicted, actual = classes[pred.argmax(0)], classes[y1]
    if predicted == actual:
      result = "Good"
      ngood = ngood + 1
    else:
      result = "Bad"
      nbad = nbad + 1
  print(f'Predicted: "{predicted}", Actual: "{actual}", Result: {result}')

print(f'{ngood} good, {nbad} bad')

a = torch.FloatTensor([[1, 2], [3, 4]]).to(device)
b = torch.matmul(a, a)
print(b)